/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.plugin.jbiplugin;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import org.apache.maven.artifact.Artifact;
import org.apache.maven.artifact.factory.ArtifactFactory;
import org.apache.maven.artifact.repository.ArtifactRepository;
import org.apache.maven.artifact.resolver.ArtifactResolver;
import org.apache.maven.artifact.versioning.InvalidVersionSpecificationException;
import org.apache.maven.artifact.versioning.VersionRange;
import org.apache.maven.model.Dependency;
import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.project.MavenProject;
import org.apache.maven.project.MavenProjectBuilder;
import org.apache.maven.project.ProjectBuildingException;
import org.apache.maven.settings.MavenSettingsBuilder;
import org.ow2.petals.jbi.descriptor.JBIDescriptorException;
import org.ow2.petals.jbi.descriptor.original.JBIDescriptorBuilder;
import org.ow2.petals.jbi.descriptor.original.generated.Jbi;

/**
 * This class is used to give the main configuration parameters of
 * all JBI goals
 *
 * @author cdeneux - Capgemini Sud
 */
public abstract class JBIAbstractMojo extends AbstractMojo {

    /*
     * Definition of all possible values of <packaging>:
     *   - jbi-component <-> binding-component or service-engine
     *   - jbi-service-unit <-> service-unit
     *   - jbi-service-assembly <-> service-assembly
     *   - jbi-shared-library <-> shared-library
     */
    protected static final String PACKAGING_COMPONENT = "jbi-component";

    protected static final String PACKAGING_SU = "jbi-service-unit";

    protected static final String PACKAGING_SA = "jbi-service-assembly";

    protected static final String PACKAGING_SL = "jbi-shared-library";

	/**
     * JBI archive name (will be suffixed by .zip).
     *
     * @parameter expression="${project.artifactId}-${project.version}"
     * @editable true
     * @required
     * @description Name of the generated jbi archive (will be suffixed by .zip).
     */
    protected String jbiName;

    /**
     * Verbose
     *
     * @parameter expression="${verbose}" default-value="false"
     * @editable true
     * @required
     * @description Boolean to switch to verbose mode
     */
    protected boolean verbose;

    /**
     * Local maven repository.
     *
     * @parameter expression="${localRepository}"
     * @required
     * @readonly
     */
    protected ArtifactRepository localRepository;

    /**
     * Output directory
     *
     * @parameter expression="${project.build.directory}"
     * @required
     */
    protected File outputDirectory;

    /**
     * The Maven Project.
     *
     * @parameter expression="${project}"
     * @required
     * @readonly
     */
    protected MavenProject project;

    /**
     * Project artifact.
     * 
     * @parameter expression="${project.artifact}"
     * @required
     * @readonly
     */
    protected Artifact projectArtifact;

    /**
     * @parameter expression="${project.packaging}"
     * @readonly
     * @required
     */
    protected String packaging;

    /**
     * JBI conf directory
     *
     * @parameter expression="${basedir}/src/main/jbi"
     * @editable true
     * @required
     * @description Path of the JBI source directory. This directory will be
     *              zipped as it is in the directory META-INF. It should contain
     *              only the file jbi.xml. Other files should be read through
     *              the classpath.
     */
    protected File jbiDirectory;

    /**
     * @component
     */
    protected ArtifactResolver artifactResolver;

    /**
     * @component
     * Needed to retrieve needed JBI archive dependencies.
     */
    protected ArtifactFactory artifactFactory;

    /**
     * @component
     * Needed to retrieve needed JBI archive dependencies.
     */
    protected MavenProjectBuilder mavenProjectBuilder;

    /**
     * @component
     */
    protected MavenSettingsBuilder settingsBuilder;

    /**
     * Read and parse the JBI descriptor of an artifact
     *
     * @param artifact JBI artifact containing the JBI descriptor. Not null.
     * @throws MojoExecutionException
     */
    final protected Jbi readJbiDescriptor(final Artifact artifact)
            throws MojoExecutionException {
        try {

            final File artifactFile = artifact.getFile();
            if (artifactFile == null) {
                throw new MojoExecutionException("Artifact file undefined (is null).");
            }
            this.debug("Read JBI descriptor from JBI archive: " + artifactFile.getAbsolutePath());
            
            final ZipFile zipFile = new ZipFile(artifactFile);
            final ZipEntry jbiDescriptorZipEntry = zipFile
                    .getEntry("META-INF/jbi.xml");
            final InputStream jbiDescriptorInputStream = zipFile
                    .getInputStream(jbiDescriptorZipEntry);

            // Load the JBI descriptor
            final Jbi jbiDescriptor = JBIDescriptorBuilder
                    .buildJavaJBIDescriptor(jbiDescriptorInputStream);

            if (jbiDescriptor == null) {
                throw new MojoExecutionException("JBI Descriptor is null");
            }

            return jbiDescriptor;

        } catch (JBIDescriptorException e) {
            throw new MojoExecutionException("Error loading JBI descriptor", e);
        } catch (IOException e) {
            throw new MojoExecutionException(
                    "Error retrieving the JBI descriptor in the JBI archive or loading the jbi xsd schema", e);
        }
    }
    
    protected void attachedJBIArchive(final Artifact artifact) {

        this.projectArtifact.setFile(
                new File(this.outputDirectory + File.separator + this.jbiName + ".zip"));

    }

    /**
     * Log the message as DEBUG one
     *
     * @param msg
     */
    protected void debug(String msg) {
        this.getLog().debug(msg);
    }

    /**
     * Log the message as INFO one
     *
     * @param msg
     */
    protected void info(String msg) {
		if (verbose) {
			this.getLog().info(msg);
		}
	}

    /**
     * Log the message as WARNING one
     *
     * @param msg
     */
    protected void warn(String msg) {
        this.getLog().warn(msg);
    }

    /**
     * Log the message as ERROR one
     *
     * @param msg
     */
    protected void error(String msg) {
		this.getLog().error(msg);
	}

    /**
     * Log the message and the exception as ERROR one
     *
     * @param msg
     * @param exception
     */
	protected void error(String msg, Throwable exception) {
		this.getLog().error(msg, exception);
	}

    /*
     * (non-Javadoc)
     * @see org.apache.maven.plugin.AbstractMojo#execute()
     */
    final public void execute() throws MojoExecutionException, MojoFailureException {
        if (this.packagingIsOkForGoal()) {
            this.executeMojo();
        }
        else {
            this.debug("The project is not a JBI artifact so the plugin skips it.");
        }
    }
    
    /**
     * Check if the current goal can be applied to the current packaging
     * @return
     */
    protected boolean packagingIsOkForGoal() {
        return PACKAGING_COMPONENT.equals(this.project.getPackaging())
                || PACKAGING_SA.equals(this.project.getPackaging())
                || PACKAGING_SL.equals(this.project.getPackaging())
                || PACKAGING_SU.equals(this.project.getPackaging());
    }

    /**
	 * Create a maven project from the given artifact
	 * @param artifact
	 * @return
	 * @throws ProjectBuildingException
	 */
	protected MavenProject buildProjectFromArtifact(Artifact artifact) throws ProjectBuildingException{
		return mavenProjectBuilder.buildFromRepository(artifact, project
				.getRemoteArtifactRepositories(),
				this.localRepository);
	}

	/**
	 * Create an Artifact from the given Dependency
	 * @param dependency MUST be non null
	 * @return the created artifact
	 * @throws InvalidVersionSpecificationException problem with the version of the dependency
	 */
	protected Artifact createDependencyArtifact(final Dependency dependency)
            throws InvalidVersionSpecificationException {
    	final Artifact dependencyArtifact = this.artifactFactory
            	.createDependencyArtifact(
            	        dependency.getGroupId(),
            			dependency.getArtifactId(),
            			VersionRange.createFromVersionSpec(
            			        dependency.getVersion()),
            			dependency.getType(),
            			dependency.getClassifier(),
            			dependency.getScope(),
            			null,
            			dependency.isOptional());

    	return dependencyArtifact;
	}

    /**
     * Retrieve artifact associated to the shared libraries needed by Maven
     * project associated to a JBI compoenent.
     * 
     * @return A shared libraries list used by the current JBI component Maven
     *         project
     * @throws MojoExecutionException
     */
    protected List<Artifact> getSharedLibraries(final MavenProject lProject) throws MojoExecutionException {
        try {
            final List<Artifact> result = new ArrayList<Artifact>();
            final List<Dependency> deps = lProject.getDependencies();
            for (final Dependency dependency : deps) {
                if (dependency.getType().equals(PACKAGING_SL)) {

                    final Artifact artifact = createDependencyArtifact(dependency);

                    result.add(artifact);
                }
            }
            return result;
        } catch (final InvalidVersionSpecificationException e) {
            throw new MojoExecutionException("Error during the artifact resolution.", e);
        }
    }

    
    /**
     * Execute the MOJO
     *
     * @throws MojoExecutionException
     * @throws MojoFailureException
     */
    abstract protected void executeMojo() throws MojoExecutionException, MojoFailureException;
}
